import parseSelector from '../util/parse-selector';
import { providerStore, componentStore, bundleStore } from '../writers';
import { Providers } from './providers';
import Module from '../classes/module';
import parsePropertyMap from '../properties/parse-property-map';
import events from '../events/events';
import directiveControllerFactory from '../util/directive-controller';
import { inputsMap } from '../properties/inputs-builder';
const TYPE = 'component';
export const Component = ({ selector, controllerAs, template, templateUrl, providers = [], inputs = [], outputs = [], pipes = [], directives = [] }) => (t) => {
    if (!selector) {
        throw new Error(`Component Decorator Error in "${t.name}": Component selector must be provided`);
    }
    let { name, type: restrict } = parseSelector(selector);
    providerStore.set('name', name, t);
    providerStore.set('type', TYPE, t);
    bundleStore.set('selector', selector, t);
    Providers(...providers)(t);
    componentStore.set('restrict', restrict, t);
    componentStore.set('scope', {}, t);
    componentStore.set('transclude', true, t);
    componentStore.set('bindToController', true, t);
    [
        ['inputs', inputs],
        ['providers', providers],
        ['directives', directives],
        ['outputs', outputs]
    ].forEach(([propName, propVal]) => {
        if (propVal !== undefined && !Array.isArray(propVal)) {
            throw new TypeError(`Component Decorator Error in "${t.name}": Component ${propName} must be an array`);
        }
    });
    let inputMap = parsePropertyMap(inputs);
    let previousInputMap = componentStore.get('inputMap', t) || {};
    componentStore.set('inputMap', Object.assign({}, previousInputMap, inputMap), t);
    if (outputs.length > 0) {
        let outputMap = parsePropertyMap(outputs) || {};
        componentStore.set('outputMap', outputMap, t);
        for (let key in outputMap) {
            events.add(outputMap[key]);
        }
    }
    if (controllerAs) {
        componentStore.set('controllerAs', controllerAs, t);
    }
    else {
        componentStore.set('controllerAs', name, t);
    }
    if (t.link) {
        componentStore.set('link', t.link, t);
    }
    if (t.compile) {
        componentStore.set('compile', t.compile, t);
    }
    View({
        selector,
        template,
        templateUrl,
        pipes,
        directives
    })(t);
};
export const View = ({ selector, template, templateUrl, pipes = [], directives = [] }) => (t) => {
    if (templateUrl) {
        componentStore.set('templateUrl', templateUrl, t);
    }
    else if (template) {
        componentStore.set('template', template, t);
    }
    else {
        throw new Error(`@Component config must include either a template or a template url for component with selector ${selector} on ${t.name}`);
    }
    Providers(...directives)(t);
    Providers(...pipes)(t);
};
Module.addProvider(TYPE, (target, name, injects, ngModule) => {
    let ddo = {};
    componentStore.forEach((val, key) => {
        ddo[key] = val;
    }, target);
    let bindProp = angular.version.minor >= 4 ? 'bindToController' : 'scope';
    ddo[bindProp] = inputsMap(ddo.inputMap);
    checkComponentConfig();
    ddo.controller = [
        '$scope', '$element', '$attrs', '$transclude', '$injector',
        function ($scope, $element, $attrs, $transclude, $injector) {
            return directiveControllerFactory(this, injects, target, ddo, $injector, {
                $scope,
                $element,
                $attrs,
                $transclude
            });
        }
    ];
    ngModule.directive(name, () => ddo);
    function createConfigErrorMessage(message) {
        return `Processing "${target.name}" in "${ngModule.name}": ${message}`;
    }
    function checkComponentConfig() {
        if (ddo.restrict !== 'E') {
            throw new Error(createConfigErrorMessage(`@Component selectors can only be elements. ` +
                `Perhaps you meant to use @Directive?`));
        }
    }
});
